(**************************************************************************)
(*                                                                        *)
(*  Decoding a PEM (Privacy Enhanced Mail) file                           *)
(*  Copyright (C) 2018   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

MODULE PEMdecode;

        (********************************************************)
        (*                                                      *)
        (*      Decoding a PEM (Privacy Enhanced Mail) file     *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            13 May 2018                     *)
        (*  Last edited:        21 May 2018                     *)
        (*  Status:             Partly written                  *)
        (*                                                      *)
        (********************************************************)


IMPORT Strings, TextIO, IOChan, BigNum;

FROM SYSTEM IMPORT CARD8;

FROM BigNum IMPORT BN;

FROM ProgramArgs IMPORT
    (* proc *)  ArgChan, IsArgPresent;

FROM FileOps IMPORT
    (* type *)  ChanId,
    (* proc *)  OpenOldFile, CloseFile, ReadLine;

FROM STextIO IMPORT
    (* proc *)  WriteChar, WriteString, WriteLn;

FROM Base64 IMPORT
    (* proc *)  Decode;

FROM LowLevel IMPORT
    (* proc *)  IAND;

(************************************************************************)

CONST
    Nul = CHR(0);
    CtrlZ = CHR(26);
    DefaultFile = "ex2.txt";

(************************************************************************)
(*                     MISCELLANOUS USEFUL PROCEDURES                   *)
(************************************************************************)

PROCEDURE WriteHexChar (N: CARD8);

    (* Write a single hexadecimal digit.  *)

    BEGIN
        IF N > 9 THEN
            WriteChar (CHR(ORD('A') + (N - 10)));
        ELSE
            WriteChar (CHR(ORD('0') + N));
        END (*IF*);
    END WriteHexChar;

(************************************************************************)

PROCEDURE WriteHex (N: CARD8);

    (* Write N in hexadecimal.  *)

    BEGIN
        WriteHexChar (N DIV 16);
        WriteHexChar (N MOD 16);
    END WriteHex;

(************************************************************************)

PROCEDURE WriteHexString (VAR (*IN*) buf: ARRAY OF CARD8;  pos, N: CARDINAL);

    (* Write N bytes starting at buf[pos].  *)

    VAR j: CARDINAL;

    BEGIN
        FOR j := 0 TO N-1 DO
            WriteHex (buf[pos]);  INC(pos);
        END (*FOR*);
    END WriteHexString;

(************************************************************************)

PROCEDURE WriteAscString (VAR (*IN*) buf: ARRAY OF CARD8;  pos, N: CARDINAL);

    (* Like WriteHexString, but we interpret the bytes as characters.  *)

    VAR j: CARDINAL;

    BEGIN
        FOR j := 0 TO N-1 DO
            WriteChar (CHR(buf[pos]));  INC(pos);
        END (*FOR*);
    END WriteAscString;

(************************************************************************)

PROCEDURE WriteCard (N: CARDINAL);

    (* Write N in decimal.  *)

    BEGIN
        IF N > 9 THEN
            WriteCard (N DIV 10);
            N := N MOD 10;
        END (*IF*);
        WriteChar (CHR(ORD('0') + N));
    END WriteCard;

(************************************************************************)
(*                              MISC UTILITIES                          *)
(************************************************************************)


(************************************************************************)
(*                                INPUT                                 *)
(************************************************************************)

PROCEDURE GetParameter (VAR (*OUT*) result: ARRAY OF CHAR);

    VAR args: IOChan.ChanId;

    BEGIN
        args := ArgChan();
        IF IsArgPresent() THEN
            TextIO.ReadString (args, result);
        ELSE
            (*result[0] := Nul;*)
            Strings.Assign (DefaultFile, result);
        END (*IF*);
    END GetParameter;

(************************************************************************)

PROCEDURE ReadPEM (VAR (*OUT*) buffer: ARRAY OF CHAR);

    (* Loads a Base64 string from a file. *)

    VAR filename, line: ARRAY [0..511] OF CHAR;
        cid: ChanId;

    BEGIN
        buffer[0] := Nul;
        GetParameter (filename);
        IF filename[0] = Nul THEN
            WriteString ("No input file was specified.");  WriteLn;
            RETURN;
        END (*IF*);
        cid := OpenOldFile (filename, FALSE, FALSE);
        LOOP
            ReadLine (cid, line);
            IF line[0] = CtrlZ THEN
                EXIT (*LOOP*);
            ELSE
                Strings.Append(line, buffer);
            END (*IF*);
        END (*LOOP*);
        CloseFile (cid);
    END ReadPEM;

(************************************************************************)
(*                        TRANSLATE PEM TO BINARY                       *)
(************************************************************************)

PROCEDURE PEMtobinary (VAR (*OUT*) binbuf: ARRAY OF CARD8): CARDINAL;

    (* Translates input file to binary, returns byte count.  *)

    CONST BufSize = 65536;

    VAR inbuff: ARRAY [0..BufSize-1] OF CHAR;

    BEGIN
        ReadPEM (inbuff);
        RETURN Decode (inbuff, binbuf);
    END PEMtobinary;

(************************************************************************)
(*                TRANSLATE BINARY ASN.1 TO HUMAN-READABLE              *)
(************************************************************************)

PROCEDURE ParseASN1 (VAR (*IN*) buffer: ARRAY OF CARD8;
                        VAR (*INOUT*) pos: CARDINAL;  N, indent: CARDINAL); FORWARD;

    (* Translates ASN.1 code starting at buffer[pos], total N bytes.    *)
    (* We update pos.                                                   *)

(************************************************************************)

PROCEDURE ParseSEQ (VAR (*IN*) buffer: ARRAY OF CARD8;
                        VAR (*INOUT*) pos: CARDINAL;  N, indent: CARDINAL);

    (* Translates a sequence.    *)

    VAR nextpos: CARDINAL;

    BEGIN
        nextpos := pos + N;
        WHILE pos < nextpos DO
            ParseASN1 (buffer, pos, N, indent);
        END (*WHILE*);
    END ParseSEQ;

(************************************************************************)

PROCEDURE ParseSET (VAR (*IN*) buffer: ARRAY OF CARD8;
                        VAR (*INOUT*) pos: CARDINAL;  N, indent: CARDINAL);

    (* Translates a set.    *)

    VAR nextpos: CARDINAL;

    BEGIN
        nextpos := pos + N;
        WHILE pos < nextpos DO
            ParseASN1 (buffer, pos, N, indent);
        END (*WHILE*);
    END ParseSET;

(************************************************************************)

PROCEDURE ParseASN1 (VAR (*IN*) buffer: ARRAY OF CARD8;
                        VAR (*INOUT*) pos: CARDINAL;  N, indent: CARDINAL);

    (* Translates ASN.1 code starting at buffer[pos], total N bytes.    *)
    (* We update pos.                                                   *)

    (********************************************************************)

    PROCEDURE Fetch(): CARD8;

        (* Gets next byte from buffer. *)

        VAR val: CARD8;

        BEGIN
            val := buffer[pos];  INC(pos);  DEC(N);
            RETURN val;
        END Fetch;

    (********************************************************************)

    PROCEDURE AppendNextByte (VAR (*INOUT*) A: BN);

        (* A := 256*A + Fetch() *)

        VAR B, C, D: BN;

        BEGIN
            B := BigNum.MakeBignum(256);
            C := BigNum.Prod (A, B);
            D := BigNum.MakeBignum (Fetch());
            BigNum.Discard (A);
            BigNum.Discard (B);
            BigNum.Discard (C);
            A := D;
        END AppendNextByte;

    (********************************************************************)

    PROCEDURE GetInt (length: BN): BN;

        (* Picks up an integer of length bytes. *)

        VAR result: BN;

        BEGIN
            result := BigNum.Zero();
            WHILE BigNum.Sign(length) > 0 DO
                AppendNextByte (result);
                BigNum.Decr(length);
            END (*WHILE*);
            RETURN result;
        END GetInt;

    (********************************************************************)

    VAR type, class, shortlength, j: CARDINAL;
        val, length: BN;
        temp: CARD8;
        constructed: BOOLEAN;

    BEGIN
        (* First byte has type information. *)

        type := Fetch();
        class := type DIV 64;
        constructed := IAND (type, 32) <> 0;
        type := IAND (type, 31);
        IF type = 31 THEN
            WriteString ("ERROR: high-tag-number form not yet supported.");
            WriteLn;
            REPEAT
                temp := Fetch();
            UNTIL IAND(temp, 128) = 0;
        END (*IF*);

        (* Second byte has length, or code to indicate long length. *)

        shortlength := Fetch();
        IF shortlength < 128 THEN
            length := BigNum.MakeBignum (shortlength);
        ELSE
            (* Long form for length. *)
            temp := shortlength - 128;  length := BigNum.Zero();
            FOR j := 0 TO temp-1 DO
                AppendNextByte (length);
            END (*FOR*);
        END (*IF*);

        (* Say what we have so far. *)

        FOR j := 1 TO indent DO
            WriteChar (' ');
        END (*FOR*);
        WriteString ("class=");  WriteCard(class);
        WriteString (", type=");  WriteCard(type);
        IF constructed THEN WriteString (" constructed") END(*IF*);
        WriteString (", length=");  BigNum.WriteBignum (length);

        shortlength := BigNum.Card(length);
        CASE type OF
            |   2:  WriteString (" INTEGER, val=");
                    val := GetInt(length);  BigNum.WriteBignum(val);  WriteLn;
                    BigNum.Discard (val);
            |   6:  WriteString (" OBJECT IDENTIFIER, val=");
                    temp := Fetch();  WriteCard(temp DIV 40);  WriteChar (' ');
                    WriteCard (temp MOD 40);
                    BigNum.Decr(length);
                    WHILE BigNum.Sign (length) > 0 DO
                        WriteChar (' ');  WriteCard(Fetch());
                        BigNum.Decr(length);
                    END (*IF*);
                    WriteLn;
            |  16:  WriteString (" SEQUENCE");  WriteLn;
                    ParseSEQ (buffer, pos, shortlength, indent+3);
            |  17:  WriteString (" SET");  WriteLn;
                    ParseSET (buffer, pos, shortlength, indent+3);
            ELSE
                    WriteLn;  WriteString ("UNKNOWN TYPE ");
                    WriteCard (type);  WriteLn;
        END (*CASE*);
        BigNum.Discard (length);

    END ParseASN1;

(************************************************************************)
(*                              MAIN PROGRAM                            *)
(************************************************************************)

PROCEDURE DoDecode;

    CONST BufSize = 65536;

    VAR binbuff: ARRAY [0..BufSize-1] OF CARD8;
        N, pos: CARDINAL;

    BEGIN
        N := PEMtobinary (binbuff);
        WriteHexString (binbuff, 0, N);  WriteLn;
        WriteCard (N);  WriteString (" bytes total");  WriteLn;
        pos := 0;
        ParseASN1 (binbuff, pos, N, 0);
    END DoDecode;

(************************************************************************)

BEGIN
    DoDecode;
END PEMdecode.

