(**************************************************************************)
(*                                                                        *)
(*  Web server input stream handler                                       *)
(*  Copyright (C) 2018   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE TLS;

        (********************************************************)
        (*                                                      *)
        (*                Transport Layer Security              *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            2 April 2018                    *)
        (*  Last edited:        23 April 2018                   *)
        (*  Status:             Partly written                  *)
        (*                                                      *)
        (********************************************************)

(************************************************************************)
(*  NOTE ABOUT VERSIONS                                                 *)
(*  Version numbering continues on from the SSL version numbers.        *)
(*  Thus we have:                                                       *)
(*          3.0 or lower    SSL versions pre-TLS                        *)
(*          3.1             TSL version 1.0                             *)
(*          3.2             TSL version 1.1                             *)
(*          3.3             TSL version 1.2                             *)
(*  I am still trying to understand the differences between these       *)
(*  versions.  Firefox for OS/2 is still using 3.1, and this does       *)
(*  make a difference to the handshake details for initial connection.  *)
(************************************************************************)


FROM STextIO IMPORT                 (* for debugging *)
    (* proc *)  WriteChar, WriteString, WriteLn;

FROM SYSTEM IMPORT CARD8, CARD16, ADR;

FROM Sockets IMPORT
    (* type *)  Socket,
    (* proc *)  recv, soclose;

FROM LowLevel IMPORT
    (* proc *)  Copy;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(************************************************************************)

TYPE
    Byte32 = ARRAY [0..31] OF CARD8;
    Byte48 = ARRAY [0..47] OF CARD8;

CONST
    Zero32 = Byte32 {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

    Zero48 = Byte48 {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

TYPE
    TLSsession = POINTER TO SessionState;

    ConnectionEnd = (server, client);
    PRFAlgorithm = (tls_prf_sha256);
    BulkCipherAlgorithm = (nocipher, rc4, tripledes, aes);
    CipherType = (stream, block, aead);
    MACAlgorithm = (null, hmac_md5, hmac_sha1, hmac_sha256, hmac_sha384,
                                 hmac_sha512);
    CompressionMethod = (nocompression, other);

    SecurityParameters =
            RECORD
                entity:                 ConnectionEnd;     (* constant? *)
                prf_algorithm:          PRFAlgorithm;      (* constant? *)
                bulk_cipher_algorithm:  BulkCipherAlgorithm;  (* used? *)
                cipher_type:            CipherType;        (* used? *)
                enc_key_length:         CARD8;             (* used? *)
                block_length:           CARD8;             (* used? *)
                fixed_iv_length:        CARD8;             (* used? *)
                record_iv_length:       CARD8;             (* used? *)
                mac_algorithm:          MACAlgorithm;      (* used? *)
                mac_length:             CARD8;             (* used? *)
                mac_key_length:         CARD8;             (* used? *)
                compression_algorithm:  CompressionMethod; (* used? *)
                master_secret:          Byte48;            (* used? *)
                client_random:          Byte32;            (* used? *)
                server_random:          Byte32;            (* used? *)
            END (*RECORD*);

    SessionState =  RECORD
                        sock:   Socket;
                        security: SecurityParameters;
                    END (*RECORD*);

    FragmentType = POINTER TO ARRAY [0..65535] OF CARD8;

    (* The standard defines a type "opaque" for both fixed-length and   *)
    (* variable-length arrays.  In the variable-length case only, such  *)
    (* an array is passed in messages as a length followed by the data. *)
    (* The number of bytes in the length depends on the maximum array   *)
    (* length; we therefore define two types, "opaque short" and        *)
    (* "opaque medium".  These are used only for variable-length byte   *)
    (* strings.  For fixed-length, we just use a fixed-length data type.*)

    OpaqueS =   RECORD
                    length: CARD8;
                    string: FragmentType;
                END (*RECORD*);

    OpaqueM =   RECORD
                    length: CARD16;
                    string: FragmentType;
                END (*RECORD*);

    (* Plain text, compressed text, and ciphertext records all have the *)
    (* same form: one-byte content type, two-byte version, two-byte     *)
    (* length, and then the data fragment itself.                       *)

    (* The content type codes are                                       *)
    (*      20      change cipher spec                                  *)
    (*      21      alert                                               *)
    (*      22      handshake                                           *)
    (*      23      application data                                    *)

    TextType =  RECORD
                    contenttype: CARD8;
                    majversion, minversion: CARD8;
                    length: CARD16;      (* BIGENDIAN *)
                    fragment: FragmentType;
                END (*RECORD*);

(************************************************************************)

PROCEDURE DoHandshake (sess: TLSsession): BOOLEAN;   FORWARD;

(************************************************************************)
(*                              DEBUGGING STUFF                         *)
(************************************************************************)

PROCEDURE WriteHexChar (N: CARD8);

    (* Write a single hexadecimal digit.  *)

    BEGIN
        IF N > 9 THEN
            WriteChar (CHR(ORD('A') + (N - 10)));
        ELSE
            WriteChar (CHR(ORD('0') + N));
        END (*IF*);
    END WriteHexChar;

(************************************************************************)

PROCEDURE WriteHex (N: CARD8);

    (* Write N in hexadecimal.  *)

    BEGIN
        WriteHexChar (N DIV 16);
        WriteHexChar (N MOD 16);
    END WriteHex;

(************************************************************************)

PROCEDURE WriteHexString (VAR (*IN*) buf: ARRAY OF CARD8;  pos, N: CARDINAL);

    (* Write N bytes starting at buf[pos].  *)

    VAR j: CARDINAL;

    BEGIN
        FOR j := 0 TO N-1 DO
            WriteHex (buf[pos]);  INC(pos);
        END (*FOR*);
    END WriteHexString;

(************************************************************************)

PROCEDURE WriteAscString (VAR (*IN*) buf: ARRAY OF CARD8;  pos, N: CARDINAL);

    (* Like WriteHexString, but we interpret the bytes as characters.  *)

    VAR j: CARDINAL;

    BEGIN
        FOR j := 0 TO N-1 DO
            WriteChar (CHR(buf[pos]));  INC(pos);
        END (*FOR*);
    END WriteAscString;

(************************************************************************)

PROCEDURE WriteCard (N: CARDINAL);

    (* Write N in decimal.  *)

    BEGIN
        IF N > 9 THEN
            WriteCard (N DIV 10);
            N := N MOD 10;
        END (*IF*);
        WriteChar (CHR(ORD('0') + N));
    END WriteCard;

(************************************************************************)

PROCEDURE BEval  (VAR (*IN*) buf: ARRAY OF CARD8;  pos, N: CARDINAL): CARDINAL;

    (* Treats the next N bytes, starting at buf[pos], as bigendian.  *)

    VAR result, k: CARDINAL;

    BEGIN
        result := 0;
        FOR k := 0 TO N-1 DO
            result := 16*result + buf[pos];
            INC (pos);
        END (*FOR*);
        RETURN result;
    END BEval;

(************************************************************************)
(*                              MISC UTILITIES                          *)
(************************************************************************)

PROCEDURE BEeval (N: CARD16): CARD16;

    (* Returns the value of a bigendian number. *)

    BEGIN
        RETURN 256*(N MOD 256) + N DIV 256;
    END BEeval;

(************************************************************************)
(*                      OPENING AND CLOSING A SESSION                   *)
(************************************************************************)

PROCEDURE OpenTLSsession (S: Socket): TLSsession;

    (* Opens a new connection.  Returns NIL if the operation failed.    *)

    VAR result: TLSsession;

    BEGIN
        NEW (result);
        result^.sock := S;
        WITH result^.security DO
            entity                  := server;
            prf_algorithm           := tls_prf_sha256;
            bulk_cipher_algorithm   := nocipher;
            cipher_type             := stream;
            enc_key_length          := 0;
            block_length            := 0;
            fixed_iv_length         := 0;
            record_iv_length        := 0;
            mac_algorithm           := null;
            mac_length              := 0;
            mac_key_length          := 0;
            compression_algorithm   := nocompression;
            master_secret           := Zero48;
            client_random           := Zero32;
            server_random           := Zero32;
        END (*WITH*);
        IF DoHandshake(result) THEN
        ELSE
            DISPOSE (result);
        END (*IF*);
        RETURN result;
    END OpenTLSsession;

(************************************************************************)

PROCEDURE CloseTLSsession (VAR (*INOUT*) sess: TLSsession);

    (* Closes the session, deletes the session data. *)

    BEGIN
        (* There are probably some formalities I have to do here, but   *)
        (* I'll come back to that later.                                *)

        soclose (sess^.sock);
        DISPOSE (sess);
    END CloseTLSsession;

(************************************************************************)
(*                              RECORD LAYER                            *)
(************************************************************************)

PROCEDURE GetFragment (sess: TLSsession;  VAR (*OUT*) buffer: TextType): BOOLEAN;

    (* Receives one record from the sender, in raw form.  It is up to   *)
    (* the caller to work out what to do with this record.              *)

    VAR buflength, size: CARDINAL;
        databuf: ARRAY [0..15] OF CARD8;

    BEGIN
        buflength := recv (sess^.sock, databuf, 5, 0);
        IF buflength = 0 THEN
            WriteString ("No data received");  WriteLn;
            buffer.fragment := NIL;
        ELSE
            WriteString ("Protocol version ");  WriteCard (databuf[1]);
            WriteChar ('.');  WriteCard (databuf[2]);  WriteLn;
            Copy (ADR(databuf), ADR(buffer), 5);
            size := BEeval (buffer.length);
            WriteString ("Fragment length is ");
            WriteCard (size);  WriteLn;
            ALLOCATE (buffer.fragment, size);
            buflength := recv (sess^.sock, buffer.fragment^, size, 0);
            RETURN TRUE;
        END (*IF*);
        RETURN FALSE;
    END GetFragment;

(************************************************************************)

PROCEDURE DiscardFragment (VAR (*INOUT*) buffer: TextType);

    BEGIN
        IF buffer.length <> 0 THEN
            DEALLOCATE (buffer.fragment, BEeval(buffer.length));
            buffer.length := 0;
            buffer.fragment := NIL;
        END (*IF*);
    END DiscardFragment;

(************************************************************************)
(*               HANDSHAKE PROTOCOL -- ClientHello EXTENSIONS           *)
(*                                                                      *)
(* Browsers that I have checked use the following extensions (hex):     *)
(*      Firefox 38.8.0    0000 0005 000A 000B 000D 0010 0023 3374 FF01  *)
(*      Firefox 59.0.2    0005 000A 000B 0010 0017 0023 0033 FF01       *)
(*                          (list possibly incomplete, there's a bug)   *)
(*                                                                      *)
(************************************************************************)

PROCEDURE Extension0 (VAR (*IN*) record: FragmentType;
                                      offset, size: CARDINAL);

    (* The server_name extension. *)

    VAR namelength: CARDINAL;
        nametype: CARD8;

    BEGIN
        WriteString ("  server_name");
        WHILE size > 0 DO

            WriteLn;

            (* In each subrecord the length is redundantly given    *)
            (* twice, so we can just skip the outer length spec.    *)

            INC (offset, 2);  DEC (size, 2);
            nametype := record^[0];
            INC (offset);  DEC (size);
            namelength := BEval(record^, offset, 2);
            INC (offset, 2);  DEC (size, 2);
            (*WriteString ("namelength=");  WriteCard(namelength);  WriteLn;*)
            WriteString ("    ");
            WriteCard (nametype);  WriteChar (' ');
            WriteAscString (record^, offset, namelength);
            INC (offset, namelength);  DEC (size, namelength);
        END (*WHILE*);
    END Extension0;

(************************************************************************)

PROCEDURE Extension13 (VAR (*IN*) record: FragmentType;
                                      offset, size: CARDINAL);

    (* The signature_algorithms extension. *)

    VAR code: CARD8;

    BEGIN
        WriteString ("  signature_algorithms");
        (*
        WriteLn;
        WriteString ("The next few bytes in the record are: ");  WriteLn;
        WriteHexString (record^, offset, 32); WriteLn;
        *)

        size := BEval(record^, offset, 2);  INC (offset, 2);
        WriteLn;

        WHILE size > 0 DO

            WriteString ("  ");

            (* Read a pair of values, one for hash algorithm and    *)
            (* one for signature algorithm.                         *)

            code := record^[offset];
            INC (offset);  DEC (size);
            CASE code OF
                    0:  WriteString ("none");
                |   1:  WriteString ("md5");
                |   2:  WriteString ("sha1");
                |   3:  WriteString ("sha224");
                |   4:  WriteString ("sha256");
                |   5:  WriteString ("sha384");
                |   6:  WriteString ("sha512");
                ELSE    WriteString ("???");
            END (*CASE*);
            WriteChar ('_');
            code := record^[offset];
            INC (offset);  DEC (size);
            CASE code OF
                    0:  WriteString ("anon");
                |   1:  WriteString ("rsa");
                |   2:  WriteString ("dsa");
                |   3:  WriteString ("ecdsa");
                ELSE    WriteString ("???");
            END (*CASE*);
        END (*WHILE*);
    END Extension13;

(************************************************************************)
(*                           HANDSHAKE PROTOCOL                         *)
(************************************************************************)

PROCEDURE ParseHandshakeExtensions (VAR (*IN*) record: FragmentType;
                                      offset, length: CARDINAL);

    (* Parses the extension fields in a ClientHello, but for now does   *)
    (* nothing with them.                                               *)

    VAR size, type: CARDINAL;

    BEGIN
        length := BEval (record^, offset, 2);
        INC (offset, 2);
        WHILE length >= 4 DO

            (* Each extension begins with four bytes: an extension  *)
            (* type and a length.                                   *)

            WriteString ("Extension type ");
            type := BEval (record^, offset, 2);
            WriteHexString (record^, offset, 2);
            INC (offset, 2);
            size := BEval (record^, offset, 2);
            INC (offset, 2);
            WriteString ("  length ");  WriteCard (size);
            IF type = 0 THEN
                Extension0 (record, offset, size);
            ELSIF type = 13 THEN
                Extension13 (record, offset, size);
            END (*IF*);
            WriteLn;
            INC (offset, size);
            DEC (length, size+4);
        END (*WHILE*);
    END ParseHandshakeExtensions;

(************************************************************************)

PROCEDURE HandleClientHello (sess: TLSsession;  VAR (*IN*) record: FragmentType;
                                      offset, length: CARDINAL): BOOLEAN;

    (* The data for this level start at record[offset].    *)

    VAR sublength, unixtime, j, k: CARDINAL;  success: BOOLEAN;
        sessionID: ARRAY [0..31] OF CARD8;
        ciphersuite: ARRAY [0..1] OF CARD8;
        suite: ARRAY [0..1] OF CARD8;

    BEGIN
        success := TRUE;

        (* The first two bytes should be the version.  Note that I  *)
        (* am updating length to show how many bytes are left in    *)
        (* this Hello message.                                      *)

        WriteString ("Protocol version ");  WriteCard (record^[offset]);
        WriteChar ('.');  WriteCard (record^[offset+1]);  WriteLn;
        DEC (length, 2);

        (* Copy client random seed. *)

        Copy (ADR(record^[offset+2]), ADR(sess^.security.client_random), 32);
        WriteString ("client_random field is ");
        WriteHexString (sess^.security.client_random, 0, 32);
        WriteLn;
        DEC (length, 32);

        (* First four bytes of seed should be Unix time. *)

        unixtime := 0;
        FOR j := 0 TO 3 DO
            unixtime := 256*unixtime + sess^.security.client_random[j];
        END (*FOR*);
        WriteString ("Unix time = ");  WriteCard (unixtime);  WriteLn;

        (* Client's session ID.  The first byte is the length.  *)

        j := offset+34;
        sublength := record^[j];  INC (j);
        DEC (length, sublength+1);
        WriteString ("SessionID length = ");  WriteCard(sublength);  WriteLn;
        k := 0;
        WHILE sublength > 0 DO
            sessionID[k] := record^[j];  INC(j);  INC(k);
            DEC (sublength);
        END (*WHILE*);
        IF k < 32 THEN
            sessionID[k] := 0;
        END (*IF*);

        (* Now we have to get a list of the cipher suites acceptable to the client. *)
        (* Each cipher suite code is two bytes long. *)

        sublength := 256*record^[j] + record^[j+1];
        INC (j, 2);
        DEC (length, sublength + 2);
        sublength := sublength DIV 2;

        WriteString ("There are ");
        WriteCard (sublength);
        WriteString (" ciphersuite records.");  WriteLn;

        WriteString ("Client will accept cipher suites:");  WriteLn;
        FOR k := 0 TO sublength-1 DO
            suite[0] := record^[j];  INC (j);
            suite[1] := record^[j];  INC (j);
            WriteString ("    ");
            WriteHex (suite[0]);  WriteChar (' ');  WriteHex (suite[1]);
            WriteLn;
        END (*FOR*);

        (* And then the compression methods. *)

        sublength := record^[j];
        WriteString ("The compression method length is ");
        WriteCard (sublength);  WriteLn;
        INC (j);
        DEC (length, sublength+1);

        WriteString ("The acceptable compression methods are: ");
        FOR k := 1 TO sublength DO
            WriteCard (record^[j]);  INC (j);
        END (*FOR*);
        WriteLn;

        IF length > 0 THEN
            WriteString ("What remains must be extensions.");  WriteLn;
            WriteString ("Length of extensions is ");
            WriteCard (length);
            WriteLn;

            ParseHandshakeExtensions (record, j, length);

            (* Dump the next few bytes in the record. *)

            (*
            WriteString ("The next few bytes in the record are: ");  WriteLn;
            WriteHexString (record^, j, 32); WriteLn;
            WriteString ("In ASCII, this is: ");  WriteLn;
            WriteAscString (record^, j, 32); WriteLn;
            sublength := 256*record^[j] + record^[j+1];
            INC (j, 2);
            DEC (length, 2);

            WriteString ("Extension length is ");  WriteCard (sublength);  WriteLn;
            exttype := 256*record^[j] + record^[j+1];
            INC (j, 2);
            DEC (length, 2);
            WriteString ("Extension type is ");  WriteCard (exttype);  WriteLn;
            IF exttype = 0 THEN
                sublength := 256*record^[j] + record^[j+1];
                INC (j, 2);
                DEC (length, 2);
                WriteString ("Server name list length is ");  WriteCard (sublength);  WriteLn;
                nametype := record^[j];  INC (j);  DEC (length);
                WriteString ("name_type is ");  WriteCard (exttype);  WriteLn;
                IF nametype = 0 THEN
                    sublength := 256*record^[j] + record^[j+1];
                    INC (j, 2);
                    DEC (length, 2);
                    WriteString ("Server name length is ");  WriteCard (sublength);  WriteLn;
                ELSE
                    WriteString ("Unsupported name_type");  WriteLn;
                END (*IF*);
            ELSE
                WriteString ("Unsupported extension type");  WriteLn;
            END (*IF*);
            *)

        END (*IF*);

        RETURN success;

    END HandleClientHello;

(************************************************************************)

PROCEDURE DoHandshake (sess: TLSsession): BOOLEAN;

    VAR type: CARD8;  success: BOOLEAN;
        j, length: CARDINAL;
        buffer: TextType;

    (* A handshake record starts with code 22 to mean "handshake", then *)
    (* we have a two-byte version code and a two-byte length, both of   *)
    (* which should have been dealt with by the caller (so I still need *)
    (* to allocate responsibilities between the layers of the protocol) *)
    (* and then we have the "fragment" field with                       *)
    (*      HandshakeType (1 byte)                                      *)
    (*      length        (3 bytes)                                     *)
    (*      content       (depends on handshake type)                   *)

    BEGIN
        success := GetFragment (sess, buffer);
        IF success THEN
            IF buffer.contenttype <> 22 THEN
                WriteString ("Received record is not a handshake record");  WriteLn;
                success := FALSE;
            ELSE
                type := buffer.fragment^[0];

                (* Interpret the three-byte length. *)

                length := 0;
                FOR j := 1 TO 3 DO
                    length := 256*length + buffer.fragment^[j];
                END (*FOR*);

                CASE type OF
                    | 1:    success := HandleClientHello (sess, buffer.fragment, 4, length);
                    |
                    ELSE
                            WriteString ("Unhandled handshake type");  WriteLn;
                            success := FALSE;
                END (*CASE*);
            END (*IF*);
        END (*IF*);
        IF NOT success THEN
            DiscardFragment (buffer);
        END (*IF*);
        RETURN success;
    END DoHandshake;

(************************************************************************)

END TLS.

