(**************************************************************************)
(*                                                                        *)
(*  Encryption library                                                    *)
(*  Copyright (C) 2017   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)


MODULE AEStest;

        (********************************************************)
        (*                                                      *)
        (*    AES: Keyed hashing for message authentication     *)
        (*                                                      *)
        (*                Test of module AES                    *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            29 October 2017                 *)
        (*  Last edited:        11 April 2018                   *)
        (*  Status:             Passes all test cases in the    *)
        (*                      standard NIST.FIPS.197          *)
        (*                                                      *)
        (*                      For CBC, passes cases 1..4 in   *)
        (*                      RFC 3602.  I have not yet had   *)
        (*                      the patience to run the         *)
        (*                      remaining cases.                *)
        (*                                                      *)
        (********************************************************)


FROM SYSTEM IMPORT CARD8, ADR;

IMPORT Strings;

FROM AES IMPORT
    (* type *)  AEScontext,
    (* proc *)  AESopen, AES_CBCopen, AESprocess, AESfinal;

FROM LowLevel IMPORT
    (* proc *)  Copy;

FROM HexData IMPORT
    (* proc *)  HexToBytes, DumpByteArray;

FROM STextIO IMPORT
    (* proc *)  WriteString, WriteLn;

(************************************************************************)

PROCEDURE DoCrypt (fullpad: BOOLEAN;  version: CARDINAL;  key, text: ARRAY OF CHAR);

    (* Does an encryption followed by a decryption. *)

    VAR M, N, length: CARDINAL;
        binkey: ARRAY [0..63] OF CARD8;
        bintext, result: ARRAY [0..63] OF CARD8;
        ctx: AEScontext;

    BEGIN
        M := HexToBytes (text, bintext);
        N := HexToBytes (key, binkey);
        WriteString (" key = ");  DumpByteArray (N, binkey);  WriteLn;
        WriteString ("text = ");  DumpByteArray (M, bintext);  WriteLn;
        ctx := AESopen (FALSE, fullpad, version, binkey);
        length := 0;
        AESprocess (ctx, M, bintext, result, length);
        AESfinal (ctx, result, length);
        WriteString (" ans = ");  DumpByteArray (length, result);  WriteLn;

        (* Now do the reverse transformation. *)

        ctx := AESopen (TRUE, fullpad, version, binkey);
        M := 0;
        AESprocess (ctx, length, result, bintext, M);
        AESfinal (ctx, bintext, M);
        WriteString ("After reverse operation,");  WriteLn;
        WriteString (" ans = ");  DumpByteArray (M, bintext);  WriteLn;
        WriteLn;

    END DoCrypt;

(************************************************************************)

PROCEDURE DoCBCgen (version, Nk: CARDINAL;  key: ARRAY OF CARD8;
                             NI: CARDINAL;  IV: ARRAY OF CARD8;
                             Nt: CARDINAL;  text: ARRAY OF CARD8);

    (* Encryption followed by decryption, for the CBC variant of AES.   *)

    CONST fullpad = TRUE;

    VAR M, length: CARDINAL;
        result: ARRAY [0..255] OF CARD8;
        ctx: AEScontext;

    BEGIN
        WriteString (" key = ");  DumpByteArray (Nk, key);  WriteLn;
        WriteString ("  IV = ");  DumpByteArray (NI, IV);  WriteLn;
        WriteString ("text = ");  DumpByteArray (Nt, text);  WriteLn;
        ctx := AES_CBCopen (FALSE, fullpad, version, key, IV);
        length := 0;
        AESprocess (ctx, Nt, text, result, length);
        AESfinal (ctx, result, length);
        WriteString (" ans = ");  DumpByteArray (length, result);  WriteLn;

        (* Now do the reverse transformation. *)

        ctx := AES_CBCopen (TRUE, fullpad, version, key, IV);
        M := 0;
        AESprocess (ctx, length, result, text, M);
        AESfinal (ctx, text, M);
        WriteString ("After decryption,");  WriteLn;
        WriteString (" ans = ");  DumpByteArray (M, text);  WriteLn;
        WriteLn;

    END DoCBCgen;

(************************************************************************)

PROCEDURE DoCBChex (version: CARDINAL;  key, IV, text: ARRAY OF CHAR);

    (* The case where key, IV, text are all hex coded. *)

    VAR Nk, NI, Nt: CARDINAL;
        binkey, binIV, bintext: ARRAY [0..255] OF CARD8;

    BEGIN
        Nk := HexToBytes (key, binkey);
        NI := HexToBytes (IV, binIV);
        Nt := HexToBytes (text, bintext);
        DoCBCgen (version, Nk, binkey, NI, binIV, Nt, bintext);
    END DoCBChex;

(************************************************************************)

PROCEDURE DoCBCtxt (version: CARDINAL;  key, IV, text: ARRAY OF CHAR);

    (* The case where the text is ASCII text, not hex coded.    *)
    (* The key and IV are still hex coded.                      *)

    VAR Nk, NI, Nt: CARDINAL;
        binkey, binIV, bintext: ARRAY [0..255] OF CARD8;

    BEGIN
        Nk := HexToBytes (key, binkey);
        NI := HexToBytes (IV, binIV);
        Nt := Strings.Length (text);
        Copy (ADR(text), ADR(bintext), Nt);
        DoCBCgen (version, Nk, binkey, NI, binIV, Nt, bintext);
    END DoCBCtxt;

(************************************************************************)

PROCEDURE DribbleInputTest (version: CARDINAL;  key, text: ARRAY OF CHAR);

    (* Like DoCrypt, but we are checking whether we can safely pass     *)
    (* the data a little bit at a time.                                 *)

    CONST fullpad = FALSE;

    VAR M, N, length, j, L1, L2: CARDINAL;
        binkey: ARRAY [0..63] OF CARD8;
        bintext, result, part2: ARRAY [0..63] OF CARD8;
        ctx: AEScontext;

    BEGIN
        M := HexToBytes (text, bintext);
        N := HexToBytes (key, binkey);
        WriteString (" key = ");  DumpByteArray (N, binkey);  WriteLn;
        WriteString ("text = ");  DumpByteArray (M, bintext);  WriteLn;

        (* Now chop bintext into parts. *)

        L1 := M DIV 3;  L2 := M - L1;
        FOR j := 0 TO L2-1 DO
            part2[j] := bintext[j+L1];
        END (*FOR*);

        ctx := AESopen (FALSE, fullpad, version, binkey);
        length := 0;
        AESprocess (ctx, L1, bintext, result, length);
        AESprocess (ctx, L2, part2, result, length);
        AESfinal (ctx, result, length);
        WriteString (" ans = ");  DumpByteArray (length, result);  WriteLn;

        (* Now do the reverse transformation. *)

        ctx := AESopen (TRUE, fullpad, version, binkey);
        M := 0;
        AESprocess (ctx, length, result, bintext, M);
        AESfinal (ctx, bintext, M);
        WriteString ("After decryption,");  WriteLn;
        WriteString (" ans = ");  DumpByteArray (M, bintext);  WriteLn;
        WriteLn;

    END DribbleInputTest;

(************************************************************************)

PROCEDURE BasicTest;

    (* Run some test vectors. *)

    CONST fullpad = FALSE;

    BEGIN
        DoCrypt (fullpad, 128, "2b 7e 15 16 28 ae d2 a6 ab f7 15 88 09 cf 4f 3c",
                            "32 43 f6 a8 88 5a 30 8d 31 31 98 a2 e0 37 07 34");
        DoCrypt (fullpad, 128, "000102030405060708090a0b0c0d0e0f",
                            "00112233445566778899aabbccddeeff");
        DoCrypt (fullpad, 192, "000102030405060708090a0b0c0d0e0f1011121314151617",
                            "00112233445566778899aabbccddeeff");
        DoCrypt (fullpad, 256, "000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f",
                            "00112233445566778899aabbccddeeff");
        DribbleInputTest (128, "2b 7e 15 16 28 ae d2 a6 ab f7 15 88 09 cf 4f 3c",
                            "32 43 f6 a8 88 5a 30 8d 31 31 98 a2 e0 37 07 34");
    END BasicTest;

(************************************************************************)

PROCEDURE CBCtests;

    (* Test vectors taken from RFC 3602. *)

    BEGIN
        DoCBCtxt (128, "06a9214036b8a15b512e03d534120006",
                 "3dafba429d9eb430b422da802c9fac41", "Single block msg");
        DoCBChex (128, "c286696d887c9aa0611bbb3e2025a45a",
                 "562e17996d093d28ddb3ba695a2e6f58",
                  "000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e1f");
        DoCBCtxt (128, "6c3ea0477630ce21a2ce334aa746c2cd",
                 "c782dc4c098c66cbd9cd27d825682c81", "This is a 48-byte message (exactly 3 AES blocks)");
        DoCBChex (128, "56e47a38c5598974bc46903dba290349",
                 "8ce82eefbea0da3c44699ed7db51b7d9",
                  "a0a1a2a3a4a5a6a7a8a9aaabacadaeaf b0b1b2b3b4b5b6b7b8b9babbbcbdbebf c0c1c2c3c4c5c6c7c8c9cacbcccdcecf d0d1d2d3d4d5d6d7d8d9dadbdcdddedf");
    END CBCtests;

(************************************************************************)

PROCEDURE PaddingTest;

    (* Tests to see whether padding is correctly added and removed. *)

    CONST fullpad = TRUE;

    BEGIN
        DoCrypt (fullpad, 128, "000102030405060708090a0b0c0d0e0f",
                               "10101010101010101010101010101010");
        DoCrypt (fullpad, 128, "000102030405060708090a0b0c0d0e0f",
                               "1010101010101010101010101010101010101010101010101010101010101010");
        DoCrypt (fullpad, 128, "000102030405060708090a0b0c0d0e0f",
                               "03030303030303030303030303");
        DoCBChex (128, "000102030405060708090a0b0c0d0e0f",
                               "8ce82eefbea0da3c44699ed7db51b7d9",
                               "1010101010101010101010101010101003030303030303030303030303");
    END PaddingTest;

(************************************************************************)

BEGIN
    BasicTest;
    CBCtests;
    PaddingTest;
END AEStest.

